import json
import os
from typing import List, Dict, Optional
from fastapi import Form, Response, UploadFile, status
from PIL import Image, UnidentifiedImageError
import yaml
import random
import numpy as np
from supervisely._utils import (
    is_production,
    is_development,
    is_debug_with_sly_net,
    rand_str,
)
from supervisely.app.fastapi.subapp import get_name_from_env
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.tag_meta import TagMeta, TagValueType

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
import supervisely.imaging.image as sly_image
import supervisely.io.fs as fs
from supervisely.sly_logger import logger
import supervisely.io.env as env

from supervisely.project.project_meta import ProjectMeta
from supervisely.app.fastapi.subapp import Application
from supervisely.app.content import StateJson, get_data_dir
from supervisely.app.fastapi.request import Request
from supervisely.api.api import Api
import supervisely.app.development as sly_app_development
from supervisely.imaging.color import get_predefined_colors
from supervisely.task.progress import Progress

try:
    from typing import Literal
except ImportError:
    # for compatibility with python 3.7
    from typing_extensions import Literal


class Inference:
    def __init__(
        self,
        model_dir: str = None,
    ):
        self._model_meta = None
        self._confidence = "confidence"
        self._app: Application = None
        self._api: Api = None
        self._task_id = None

        self._model_dir = None
        self._local_dir = None
        if model_dir is not None:
            self._prepare_model_directory(model_dir)

        self._headless = True

    def _prepare_model_directory(self, model_dir):
        self._model_dir = model_dir
        self._local_dir = None
        if fs.is_on_agent(self._model_dir) or is_production():
            team_id = env.team_id()
            logger.info(f"Remote model directory in Team Files: {self._model_dir}")
            self._local_dir = os.path.join(get_data_dir(), "model")
            logger.info(f"Local model directory: {self._local_dir}")

            if fs.dir_exists(self._local_dir):
                # only during debug, has no effect in production
                pass
            else:
                sizeb = self.api.file.get_directory_size(team_id, self._model_dir)
                progress = Progress(
                    "Downloading model directory ...",
                    total_cnt=sizeb,
                    is_size=True,
                )
                self.api.file.download_directory(
                    team_id,
                    self._model_dir,
                    self._local_dir,
                    progress.iters_done_report,
                )
                print(f"📥 Model directory has been successfully downloaded from Team Files")
        else:
            self._model_dir = os.path.abspath(self._model_dir)
            self._local_dir = self._model_dir
            print(f"Model directory: {self._local_dir}")

    def _prepare_device(self, device):
        if device is None:
            try:
                import torch

                device = "cuda" if torch.cuda.is_available() else "cpu"
            except Exception as e:
                logger.warn(
                    f"Device auto detection failed, set to default 'cpu', reason: {repr(e)}"
                )
                device = "cpu"

    def _get_templates_dir(self):
        return None
        # raise NotImplementedError("Have to be implemented in child class")

    def _get_layout(self):
        return None
        # raise NotImplementedError("Have to be implemented in child class")

    def load_on_device(
        device: Literal["cpu", "cuda", "cuda:0", "cuda:1", "cuda:2", "cuda:3"] = "cpu"
    ):
        raise NotImplementedError("Have to be implemented in child class after inheritance")

    def get_classes(self) -> List[str]:
        raise NotImplementedError("Have to be implemented in child class after inheritance")

    def get_info(self) -> dict:
        return {
            "model name": get_name_from_env(default="Neural Network Serving"),
            "model dir": self.model_dir,
            "number of classes": len(self.get_classes()),
        }

    @property
    def model_dir(self):
        return self._local_dir

    @property
    def api(self) -> Api:
        if self._api is None:
            self._api = Api()
        return self._api

    def _get_obj_class_shape(self):
        raise NotImplementedError("Have to be implemented in child class")

    @property
    def model_meta(self) -> ProjectMeta:
        if self._model_meta is None:
            colors = get_predefined_colors(len(self.get_classes()))
            classes = []
            for name, rgb in zip(self.get_classes(), colors):
                classes.append(ObjClass(name, self._get_obj_class_shape(), rgb))
            self._model_meta = ProjectMeta(classes)
            self._get_confidence_tag_meta()
        return self._model_meta

    
    @property
    def task_id(self) -> int:
        return self._task_id

    def _get_confidence_tag_meta(self):
        tag_meta = self.model_meta.get_tag_meta(self._confidence)
        if tag_meta is None:
            tag_meta = TagMeta(self._confidence, TagValueType.ANY_NUMBER)
            self._model_meta = self._model_meta.add_tag_meta(tag_meta)
        return tag_meta

    def _create_label(self, dto) -> Label:
        raise NotImplementedError("Have to be implemented in child class")

    def predict(self, image_path: str):
        raise NotImplementedError("Have to be implemented in child class")

    def predict_annotation(self, image_path: str) -> Annotation:
        raise NotImplementedError("Have to be implemented in child class")

    def visualize(self, predictions: List, image_path: str, vis_path: str):
        raise NotImplementedError("Have to be implemented in child class")

    def _predictions_to_annotation(self, image_path: str, predictions: List) -> Annotation:
        labels = []
        for prediction in predictions:
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = Annotation.from_img_path(image_path)
        ann = ann.add_labels(labels)
        return ann

    def _get_custom_inference_settings() -> str:  # in yaml format
        return ""

    def inference_image_path(
        self, image_path, project_meta: ProjectMeta, state: Dict, settings: Dict = None
    ):
        raise NotImplementedError()

    def _get_custom_inference_settings_dict(self) -> dict:
        return yaml.safe_load(self._get_custom_inference_settings())

    def get_inference_settings(self, state: dict):
        settings = state.get("settings", {})
        for key, value in self._get_custom_inference_settings_dict().items():
            if key not in settings:
                logger.warn(
                    f"Field {key} not found in inference settings. Use default value {value}"
                )
                settings[key] = value
        return settings

    @property
    def app(self) -> Application:
        return self._app

    def inference_image(self, state: dict, file: UploadFile):
        logger.debug("Input state", extra={"state": state})
        settings = self.get_inference_settings(state)
        image_path = os.path.join(get_data_dir(), f"{rand_str(10)}_{file.filename}")
        image_np = sly_image.read_bytes(file.file.read())
        sly_image.write(image_path, image_np)
        ann_json = self.inference_image_path(
            image_path=image_path,
            project_meta=self.model_meta,
            state=state,
            settings=settings,
        )
        fs.silent_remove(image_path)
        return ann_json

    def inference_batch(self, state: dict, files: List[UploadFile]):
        logger.debug("Input state", extra={"state": state})
        paths = []
        temp_dir = os.path.join(get_data_dir(), rand_str(10))
        fs.mkdir(temp_dir)
        for file in files:
            image_path = os.path.join(temp_dir, f"{rand_str(10)}_{file.filename}")
            image_np = sly_image.read_bytes(file.file.read())
            sly_image.write(image_path, image_np)
            paths.append(image_path)
        results = self.inference_images_dir(paths, state)
        fs.remove_dir(temp_dir)
        return results

    def inference_batch_ids(self, api: Api, state: dict):
        ids = state["batch_ids"]
        infos = api.image.get_info_by_id_batch(ids)
        paths = []
        temp_dir = os.path.join(get_data_dir(), rand_str(10))
        fs.mkdir(temp_dir)
        for info in infos:
            paths.append(os.path.join(temp_dir, f"{rand_str(10)}_{info.name}"))
        api.image.download_paths(infos[0].dataset_id, ids, paths)
        results = self.inference_images_dir(paths, state)
        fs.remove_dir(temp_dir)
        return results

    def inference_images_dir(self, img_paths: List[str], state: Dict):
        settings = self.get_inference_settings(state)
        annotations = []
        for image_path in img_paths:
            ann_json = self.inference_image_path(
                image_path=image_path,
                project_meta=self.model_meta,
                state=state,
                settings=settings,
            )
            annotations.append(ann_json)
        return annotations

    def inference_image_id(self, api: Api, state: dict):
        logger.debug("Input state", extra={"state": state})
        settings = self.get_inference_settings(state)
        image_id = state["image_id"]
        image_info = api.image.get_info_by_id(image_id)
        image_path = os.path.join(get_data_dir(), f"{rand_str(10)}_{image_info.name}")
        api.image.download_path(image_id, image_path)
        ann_json = self.inference_image_path(
            image_path=image_path,
            project_meta=self.model_meta,
            state=state,
            settings=settings,
        )
        fs.silent_remove(image_path)
        return ann_json

    def inference_image_url(self, api: Api, state: dict):
        logger.debug("Input data", extra={"state": state})
        settings = self.get_inference_settings(state)
        image_url = state["image_url"]
        ext = fs.get_file_ext(image_url)
        if ext == "":
            ext = ".jpg"
        image_path = os.path.join(get_data_dir(), rand_str(15) + ext)
        fs.download(image_url, image_path)
        ann_json = self.inference_image_path(
            image_path=image_path,
            project_meta=self.model_meta,
            state=state,
            settings=settings,
        )
        fs.silent_remove(image_path)
        return ann_json
    
    def inference_video_id(self, api: Api, state: dict):
        from supervisely.nn.inference.video_inference import InferenceVideoInterface
        logger.debug("Input data", extra={"state": state})
        video_info = api.video.get_info_by_id(state['videoId'])

        video_images_path = os.path.join(get_data_dir(), rand_str(15))
        inf_video_interface = InferenceVideoInterface(
            api=api,
            start_frame_index=state.get('startFrameIndex', 0),
            frames_count=state.get('framesCount', video_info.frames_count - 1),
            frames_direction=state.get('framesDirection', 'forward'),
            video_info=video_info,
            imgs_dir=video_images_path,
        )

        inf_video_interface.download_frames()
        settings = self.get_inference_settings(state)
        anns_json = []
        for image_path in inf_video_interface.images_paths:
            ann_json = self.inference_image_path(
                image_path=image_path, 
                project_meta=self.model_meta,
                state=state,
                settings=settings,
            )
            anns_json.append(ann_json)
        fs.remove_dir(video_images_path)
        return anns_json

    def serve(self):
        if is_debug_with_sly_net():
            # advanced debug for Supervisely Team
            logger.warn(
                "Serving is running in advanced development mode with Supervisely VPN Network"
            )
            team_id = env.team_id()
            # sly_app_development.supervisely_vpn_network(action="down") # for debug
            sly_app_development.supervisely_vpn_network(action="up")
            task = sly_app_development.create_debug_task(team_id, port="8000")
            self._task_id = task["id"]
        else:
            self._task_id = env.task_id()

        # headless=self._headless,
        self._app = Application(layout=self._get_layout(), templates_dir=self._get_templates_dir())
        server = self._app.get_server()

        @server.post(f"/get_session_info")
        def get_session_info():
            return self.get_info()

        @server.post("/get_custom_inference_settings")
        def get_custom_inference_settings():
            settings = self._get_custom_inference_settings()
            return {"settings": settings}

        @server.post("/get_output_classes_and_tags")
        def get_output_classes_and_tags():
            return self.model_meta.to_json()

        @server.post("/inference_image_id")
        def inference_image_id(request: Request):
            return self.inference_image_id(request.api, request.state)

        @server.post("/inference_image_url")
        def inference_image_url(request: Request):
            return self.inference_image_url(request.api, request.state)

        @server.post("/inference_batch_ids")
        def inference_batch_ids(request: Request):
            return self.inference_batch_ids(request.api, request.state)

        @server.post("/inference_video_id")
        def inference_video_id(request: Request):
            return {'ann': self.inference_video_id(request.api, request.state)}

        @server.post("/inference_image")
        def inference_image(response: Response, files: List[UploadFile], settings: str = Form("{}")):
            if len(files) != 1:
                response.status_code = status.HTTP_400_BAD_REQUEST
                return f'Only one file expected but got {len(files)}'
            try:
                state = json.loads(settings)
                if type(state) != dict:
                    response.status_code = status.HTTP_400_BAD_REQUEST
                    return 'Settings is not json object'
                return self.inference_image(state, files[0])
            except (json.decoder.JSONDecodeError, TypeError) as e:
                response.status_code = status.HTTP_400_BAD_REQUEST
                return f'Cannot decode settings: {e}'
            except sly_image.UnsupportedImageFormat:
                response.status_code = status.HTTP_400_BAD_REQUEST
                return f'File has unsupported format. Supported formats: {sly_image.SUPPORTED_IMG_EXTS}'

        @server.post("/inference_batch")
        def inference_batch(response: Response, files: List[UploadFile], settings: str = Form("{}")):
            try:
                state = json.loads(settings)
                if type(state) != dict:
                    response.status_code = status.HTTP_400_BAD_REQUEST
                    return 'Settings is not json object'
                return self.inference_batch(state, files)
            except (json.decoder.JSONDecodeError, TypeError) as e:
                response.status_code = status.HTTP_400_BAD_REQUEST
                return f'Cannot decode settings: {e}'
            except sly_image.UnsupportedImageFormat:
                response.status_code = status.HTTP_400_BAD_REQUEST
                return f'File has unsupported format. Supported formats: {sly_image.SUPPORTED_IMG_EXTS}'



from typing import Dict, List, Any
from pathlib import Path
import os
from supervisely.app.widgets.widget import Widget
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.prediction_dto import PredictionBBox
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
import supervisely.imaging.image as sly_image
from supervisely.sly_logger import logger
from supervisely.nn.inference.inference import Inference
from supervisely.annotation.annotation import Annotation
from supervisely.decorators.inference import (
    process_image_roi,
    process_image_sliding_window,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.progress import Progress


class ObjectDetection(Inference):
    def _get_templates_dir(self):
        # template_dir = os.path.join(
        #     Path(__file__).parent.absolute(), "dashboard/templates"
        # )
        # return template_dir
        return None

    def _get_layout(self) -> Widget:
        return None
        # import supervisely.nn.inference.instance_segmentation.dashboard.main_ui as main_ui
        # return main_ui.menu

    def get_info(self) -> dict:
        info = super().get_info()
        info["task type"] = "object detection"
        info["sliding_window_support"] = "basic"  # or "advanced" in the future
        return info

    def _get_obj_class_shape(self):
        return Rectangle

    def _create_label(self, dto: PredictionBBox):
        obj_class = self.model_meta.get_obj_class(dto.class_name)
        if obj_class is None:
            raise KeyError(
                f"Class {dto.class_name} not found in model classes {self.get_classes()}"
            )
        geometry = Rectangle(*dto.bbox_tlbr)
        tags = []
        if dto.score is not None:
            tags.append(Tag(self._get_confidence_tag_meta(), dto.score))
        label = Label(geometry, obj_class, tags)
        return label

    def _get_custom_inference_settings(self) -> str:  # in yaml format
        settings = """confidence_threshold: 0.5"""
        return settings

    def predict(self, image_path: str, confidence_threshold: float) -> List[PredictionBBox]:
        raise NotImplementedError("Have to be implemented in child class")

    def predict_annotation(self, image_path: str, settings: Dict[str, Any]) -> Annotation:
        predictions = self.predict(image_path, settings)
        return self._predictions_to_annotation(image_path, predictions)

    @process_image_roi
    def inference_image_path(
        self,
        image_path: str,
        project_meta: ProjectMeta,
        state: Dict,
        settings: Dict = None,
    ):
        if settings is None:
            settings = self.get_inference_settings(state)
        logger.debug("Input path", extra={"path": image_path})
        ann = self.predict_annotation(
            image_path, settings=settings
        )
        if isinstance(ann, Annotation):
            return ann.to_json()
        else:
            return ann

    def visualize(self, predictions: List[PredictionBBox], image_path: str, vis_path: str):
        image = sly_image.read(image_path)
        ann = self._predictions_to_annotation(image_path, predictions)
        ann.draw_pretty(bitmap=image, output_path=vis_path, fill_rectangles=False)

    def serve(self):
        # import supervisely.nn.inference.instance_segmentation.dashboard.main_ui as main_ui
        # import supervisely.nn.inference.instance_segmentation.dashboard.deploy_ui as deploy_ui

        # @deploy_ui.deploy_btn.click
        # def deploy_model():
        # device = deploy_ui.device.get_value()
        # self.load_on_device(self._device)
        # print(f"✅ Model has been successfully loaded on {self._device.upper()} device")
        Progress("Deploying model ...", 1)
        super().serve()
        Progress("Model deployed", 1).iter_done_report()



# coding: utf-8

# docs
from __future__ import annotations
import cv2
import numpy as np
from typing import List, Tuple, Dict, Optional
import supervisely as sly


from supervisely.geometry.constants import EXTERIOR, INTERIOR, POINTS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_location import PointLocation, points_to_row_col_list
from supervisely.geometry import validation


class Rectangle(Geometry):
    """
    Rectangle geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Rectangle<Rectangle>` class object is immutable.

    :param top: Minimal vertical value of Rectangle object.
    :type top: int or float
    :param left: Minimal horizontal value of Rectangle object.
    :type left: int or float
    :param bottom: Maximal vertical value of Rectangle object.
    :type bottom: int or float
    :param right: Maximal vertical value of Rectangle object.
    :type right: int or float
    :param sly_id: Rectangle ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Rectangle belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Rectangle.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Rectangle was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Rectangle was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`. Rectangle top argument must have less or equal value then bottom, left argument must have less or equal value then right

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        top = 100
        left = 100
        bottom = 700
        right = 900
        figure = sly.Rectangle(top, left, bottom, right)
    """
    @staticmethod
    def geometry_name():
        """
        """
        return 'rectangle'

    def __init__(self, top: int, left: int, bottom: int, right: int,
                 sly_id: Optional[int] = None, class_id: Optional[int] = None, labeler_login: Optional[int] = None,
                 updated_at: Optional[str] = None, created_at: Optional[str] = None):

        if top > bottom:
            raise ValueError('Rectangle "top" argument must have less or equal value then "bottom"!')

        if left > right:
            raise ValueError('Rectangle "left" argument must have less or equal value then "right"!')

        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at,
                         created_at=created_at)

        self._points = [PointLocation(row=top, col=left), PointLocation(row=bottom, col=right)]

    def to_json(self) -> Dict:
        """
        Convert the Rectangle to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            figure_json = figure.to_json()
            print(figure_json)
            # Output: {
            #    "points": {
            #        "exterior": [
            #            [100, 100],
            #            [900, 700]
            #        ],
            #        "interior": []
            #    }
            # }
        """
        packed_obj = {
            POINTS: {
                EXTERIOR: points_to_row_col_list(self._points, flip_row_col_order=True),
                INTERIOR: []
            }
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @classmethod
    def from_json(cls, data: Dict) -> Rectangle:
        """
        Convert a json dict to Rectangle. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Rectangle in json format as a dict.
        :type data: dict
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_json = {
                "points": {
                    "exterior": [
                        [100, 100],
                        [900, 700]
                    ],
                    "interior": []
                }
            }
            figure = sly.Rectangle.from_json(figure_json)
        """
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)

        exterior = data[POINTS][EXTERIOR]
        if len(exterior) != 2:
            raise ValueError('"exterior" field must contain exactly two points to create Rectangle object.')
        [top, bottom] = sorted([exterior[0][1], exterior[1][1]])
        [left, right] = sorted([exterior[0][0], exterior[1][0]])
        return cls(top=top, left=left, bottom=bottom, right=right,
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def crop(self, other: Rectangle) -> List[Rectangle]:
        """
        Crops current Rectangle.

        :param rect: Rectangle object for crop.
        :type rect: Rectangle
        :return: List of Rectangle objects
        :rtype: :class:`List[Rectangle]<Rectangle>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            crop_figures = figure.crop(sly.Rectangle(0, 0, 300, 350))
        """
        top = max(self.top, other.top)
        left = max(self.left, other.left)
        bottom = min(self.bottom, other.bottom)
        right = min(self.right, other.right)
        is_valid = (bottom >= top) and (left <= right)
        return [Rectangle(top=top, left=left, bottom=bottom, right=right)] if is_valid else []

    def _transform(self, transform_fn):
        """
        """
        transformed_corners = [transform_fn(p) for p in self.corners]
        rows, cols = zip(*points_to_row_col_list(transformed_corners))
        return Rectangle(top=round(min(rows)), left=round(min(cols)), bottom=round(max(rows)), right=round(max(cols)))

    @property
    def corners(self) -> List[PointLocation, PointLocation, PointLocation, PointLocation]:
        """
        Get list of Rectangle corners.

        :return: List of PointLocation objects
        :rtype: :class:`List[PointLocation]<supervisely.geometry.point_location.PointLocation>`

        :Usage Example:

         .. code-block:: python

            corners = figure.corners
            for corner in corners:
                print(corner.row, corner.col)
            # Output:
            # 100 100
            # 100 900
            # 700 900
            # 700 100
        """
        return [PointLocation(row=self.top, col=self.left), PointLocation(row=self.top, col=self.right),
                PointLocation(row=self.bottom, col=self.right), PointLocation(row=self.bottom, col=self.left)]

    def rotate(self, rotator: sly.geometry.image_rotator.ImageRotator) -> Rectangle:
        """
        Rotates current Rectangle.

        :param rotator: ImageRotator object for rotation.
        :type rotator: ImageRotator
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            from supervisely.geometry.image_rotator import ImageRotator

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            height, width = 300, 400
            rotator = ImageRotator((height, width), 25)
            rotate_figure = figure.rotate(rotator)
        """
        return self._transform(lambda p: rotator.transform_point(p))

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> Rectangle:
        """
        Resizes current Rectangle.

        :param in_size: Input image size (height, width) to which belongs Rectangle.
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) to which belongs Rectangle.
        :type out_size: Tuple[int, int]
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            in_height, in_width = 300, 400
            out_height, out_width = 600, 800
            resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor: float) -> Rectangle:
        """
        Scales current Rectangle.

        :param factor: Scale parameter.
        :type factor: float
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            scale_figure = figure.scale(0.75)
        """
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow: int, dcol: int) -> Rectangle:
        """
        Translates current Rectangle.

        :param drow: Horizontal shift.
        :type drow: int
        :param dcol: Vertical shift.
        :type dcol: int
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            translate_figure = figure.translate(150, 250)
        """
        return self._transform(lambda p: p.translate(drow, dcol))

    def fliplr(self, img_size: Tuple[int, int]) -> Rectangle:
        """
        Flips current Rectangle in horizontal.

        :param img_size: Input image size (height, width) to which belongs Rectangle.
        :type img_size: Tuple[int, int]
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            height, width = 300, 400
            fliplr_figure = figure.fliplr((height, width))
        """
        img_width = img_size[1]
        return Rectangle(top=self.top, left=(img_width - self.right), bottom=self.bottom, right=(img_width - self.left))

    def flipud(self, img_size: Tuple[int, int]) -> Rectangle:
        """
        Flips current Rectangle in vertical.

        :param img_size: Input image size (height, width) to which belongs Rectangle.
        :type img_size: Tuple[int, int]
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            height, width = 300, 400
            flipud_figure = figure.flipud((height, width))
        """
        img_height = img_size[0]
        return Rectangle(top=(img_height - self.bottom), left=self.left, bottom=(img_height - self.top),
                         right=self.right)

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """
        """
        self._draw_contour_impl(bitmap, color, thickness=cv2.FILLED, config=config)  # due to cv2

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """
        """
        cv2.rectangle(bitmap, pt1=(self.left, self.top), pt2=(self.right, self.bottom), color=color,
                      thickness=thickness)

    def to_bbox(self) -> Rectangle:
        """
        Makes a copy of Rectangle.

        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
            new_figure = figure.to_bbox()
        """
        return self.clone()

    @property
    def area(self) -> float:
        """
        Rectangle area.

        :return: Area of current Rectangle object
        :rtype: :class:`float`

        :Usage Example:

         .. code-block:: python

            print(figure.area)
            # Output: 7288.0
        """
        return float(self.width * self.height)

    @classmethod
    def from_array(cls, arr: np.ndarray) -> Rectangle:
        """
        Create Rectangle with given array shape.

        :param arr: Numpy array.
        :type arr: np.ndarray
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            np_array = np.zeros((300, 400))
            figure_from_np = sly.Rectangle.from_array(np_array)
        """
        return cls(top=0, left=0, bottom=arr.shape[0] - 1, right=arr.shape[1] - 1)

    # TODO re-evaluate whether we need this, looks trivial.
    @classmethod
    def from_size(cls, size: Tuple[int, int]) -> Rectangle:
        """
        Create Rectangle with given size shape.

        :param size: Input size.
        :type size: Tuple[int, int]
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            size = (300, 400)
            figure_from_size = sly.Rectangle.from_size(size)
        """
        return cls(0, 0, size[0] - 1, size[1] - 1)

    @classmethod
    def from_geometries_list(cls, geometries: List[sly.geometry.geometry]) -> Rectangle:
        """
        Create Rectangle from given geometry objects.

        :param geometries: List of geometry type objects: :class:`Bitmap<supervisely.geometry.bitmap.Bitmap>`, :class:`Cuboid<supervisely.geometry.cuboid.Cuboid>`, :class:`Point<supervisely.geometry.point.Point>`, :class:`Polygon<supervisely.geometry.polygon.Polygon>`, :class:`Polyline<supervisely.geometry.polyline.Polyline>`, :class:`Rectangle<Rectangle>`, :class:`Graph<supervisely.geometry.graph.GraphNodes>`.
        :type geometries: List[Geometry]
        :return: Rectangle object
        :rtype: :class:`Rectangle<Rectangle>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            geom_objs = [sly.Point(100, 200), sly.Polyline([sly.PointLocation(730, 2104), sly.PointLocation(2479, 402)])]
            figure_from_geom_objs = sly.Rectangle.from_geometries_list(geom_objs)
        """
        bboxes = [g.to_bbox() for g in geometries]
        top = min(bbox.top for bbox in bboxes)
        left = min(bbox.left for bbox in bboxes)
        bottom = max(bbox.bottom for bbox in bboxes)
        right = max(bbox.right for bbox in bboxes)
        return cls(top=top, left=left, bottom=bottom, right=right)

    @property
    def left(self) -> int:
        """
        Minimal horizontal value of Rectangle.

        :return: Minimal horizontal value
        :rtype: :class:`int`

        :Usage Example:

         .. code-block:: python

            print(figure.left)
            # Output: 100
        """
        return self._points[0].col

    @property
    def right(self) -> int:
        """
        Maximal horizontal value of Rectangle.

        :return: Maximal horizontal value
        :rtype: :class:`int`

        :Usage Example:

         .. code-block:: python

            print(figure.right)
            # Output: 900
        """
        return self._points[1].col

    @property
    def top(self) -> int:
        """
        Minimal vertical value of Rectangle.

        :return: Minimal vertical value
        :rtype: :class:`int`

        :Usage Example:

         .. code-block:: python

            print(rectangle.top)
            # Output: 100
        """
        return self._points[0].row

    @property
    def bottom(self) -> int:
        """
        Maximal vertical value of Rectangle.

        :return: Maximal vertical value
        :rtype: :class:`int`

        :Usage Example:

         .. code-block:: python

            print(figure.bottom)
            # Output: 700
        """
        return self._points[1].row

    @property
    def center(self) -> PointLocation:
        """
        Center of Rectangle.

        :return: PointLocation object
        :rtype: :class:`PointLocation<supervisely.geometry.point_location.PointLocation>`

        :Usage Example:

         .. code-block:: python

            center = figure.center()
        """
        return PointLocation(row=(self.top + self.bottom) // 2, col=(self.left + self.right) // 2)

    @property
    def width(self) -> int:
        """
        Width of Rectangle.

        :return: Width
        :rtype: :class:`int`

        :Usage Example:

         .. code-block:: python

            print(figure.width)
            # Output: 801
        """
        return self.right - self.left + 1

    @property
    def height(self) -> int:
        """
        Height of Rectangle

        :return: Height
        :rtype: :class:`int`

        :Usage Example:

         .. code-block:: python

            print(figure.height)
            # Output: 601
        """
        return self.bottom - self.top + 1

    def contains(self, rect: Rectangle) -> bool:
        """
        Checks if Rectangle contains a given Rectangle object.

        :param rect: Rectangle object.
        :type rect: Rectangle
        :return: True if Rectangle contains given Rectangle object, otherwise False
        :rtype: :class:`bool`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            rect = sly.Rectangle(200, 250, 400, 500))
            print(figure.contains(rect))
            # Output: True
        """
        return (self.left <= rect.left and
                self.right >= rect.right and
                self.top <= rect.top and
                self.bottom >= rect.bottom)

    def contains_point_location(self, pt: PointLocation) -> bool:
        """
        Checks if Rectangle contains a given PointLocation object.

        :param pt: PointLocation object.
        :type pt: PointLocation
        :return: True if Rectangle contains given PointLocation object, otherwise False
        :rtype: :class:`bool`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            pt = sly.PointLocation(250, 300))
            print(figure.contains_point_location(pt))
            # Output: True
        """
        return (self.left <= pt.col <= self.right) and (self.top <= pt.row <= self.bottom)

    def to_size(self) -> Tuple[int, int]:
        """
        Height and width of Rectangle.

        :return: Height and width of Rectangle object
        :rtype: :class:`Tuple[int, int]`

        :Usage Example:

         .. code-block:: python

            height, width = figure.to_size()
            print(height, width)
            # Output: 700 900
        """
        return self.height, self.width

    def get_cropped_numpy_slice(self, data: np.ndarray) -> np.ndarray:
        """
        Slice of given numpy array with Rectangle.

        :param data: Numpy array.
        :type data: np.ndarray
        :return: Sliced numpy array
        :rtype: :class:`np.ndarray<np.ndarray>`

        :Usage Example:

         .. code-block:: python

            np_slice = np.zeros((200, 500))
            mask_slice = figure.get_cropped_numpy_slice(np_slice)
            print(mask_slice.shape)
            # Output: (199, 499)
        """
        return data[self.top:(self.bottom+1), self.left:(self.right+1), ...]

    def intersects_with(self, rect: Rectangle) -> bool:
        """
        Checks intersects Rectangle with given Rectangle object or not.

        :param rect: Rectangle object.
        :type rect: Rectangle
        :return: True if given Rectangle object intersects with Rectangle, otherwise False
        :rtype: :class:`bool`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            rect = sly.Rectangle(90, 90, 400, 500)
            print(figure.intersects_with(rect))
            # Output: True
        """
        if self.left > rect.right or self.right < rect.left:
            return False
        if self.top > rect.bottom or self.bottom < rect.top:
            return False
        return True

    @classmethod
    def allowed_transforms(cls):
        """
        """
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.polygon import Polygon
        return [AnyGeometry, Bitmap, Polygon]



from typing import Optional, List, Dict, Any

from supervisely.app.widgets import Widget
from supervisely.app.widgets.widget import Disableable
import supervisely as sly
from supervisely.geometry.geometry import Geometry
from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.sly_logger import logger
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.pointcloud import Pointcloud
from supervisely.geometry.point_3d import Point3d

type_to_zmdi_icon = {
    sly.AnyGeometry: "zmdi zmdi-shape",
    sly.Rectangle: "zmdi zmdi-crop-din",  # "zmdi zmdi-square-o"
    sly.Polygon: "icons8-polygon",  # "zmdi zmdi-edit"
    sly.Bitmap: "zmdi zmdi-brush",
    sly.Polyline: "zmdi zmdi-gesture",
    sly.Point: "zmdi zmdi-dot-circle-alt",
    sly.Cuboid: "zmdi zmdi-ungroup",  #
    Cuboid3d: "zmdi zmdi-codepen",
    Pointcloud: "zmdi zmdi-cloud-outline",  #  # "zmdi zmdi-border-clear"
    sly.MultichannelBitmap: "zmdi zmdi-layers",  # "zmdi zmdi-collection-item"
    Point3d: "zmdi zmdi-filter-center-focus",  # "zmdi zmdi-select-all"
}


class ClassesTable(Widget):
    class Routes:
        CLASS_SELECTED = "class_selected_cb"

    def __init__(
        self,
        project_meta: Optional[sly.ProjectMeta] = None,
        project_id: Optional[int] = None,
        project_fs: Optional[sly.Project] = None,
        allowed_types: Optional[List[Geometry]] = None,
        selectable: Optional[bool] = True,
        disabled: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        if project_id is not None and project_fs is not None:
            raise ValueError(
                "You can not provide both project_id and project_fs parameters to Classes Table widget."
            )
        self._table_data = []
        self._columns = []
        self._changes_handled = False
        self._global_checkbox = False
        self._checkboxes = []
        self._selectable = selectable
        self._disabled = disabled
        self._loading = False
        self._allowed_types = allowed_types if allowed_types is not None else []
        if project_id is not None:
            self._api = sly.Api()
        else:
            self._api = None
        self._project_id = project_id
        if project_id is not None:
            if project_meta is not None:
                logger.warn(
                    "Both parameters project_id and project_meta were provided to ClassesTable widget. Project meta classes taken from remote project and project_meta parameter is ignored."
                )
            project_meta = sly.ProjectMeta.from_json(
                self._api.project.get_meta(project_id)
            )
        self._project_fs = project_fs
        if project_fs is not None:
            if project_meta is not None:
                logger.warn(
                    "Both parameters project_fs and project_meta were provided to ClassesTable widget. Project meta classes taken from project_fs.meta and project_meta parameter is ignored."
                )
            project_meta = project_fs.meta
        if project_meta is not None:
            self._update_meta(project_meta=project_meta)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def value_changed(self, func):
        route_path = self.get_route_path(ClassesTable.Routes.CLASS_SELECTED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_selected_classes()
            func(res)

        return _value_changed

    def _update_meta(self, project_meta: sly.ProjectMeta) -> None:
        columns = ["class", "shape"]
        stats = None
        data_to_show = []
        for obj_class in project_meta.obj_classes:
            if (
                self._allowed_types is None
                or obj_class.geometry_type not in self._allowed_types
            ):
                data_to_show.append(obj_class.to_json())

        if self._project_id is not None:
            stats = self._api.project.get_stats(self._project_id)
            project_info = self._api.project.get_info_by_id(self._project_id)
            if project_info.type == str(sly.ProjectType.IMAGES):
                columns.extend(["images count", "labels count"])
            elif project_info.type == str(sly.ProjectType.VIDEOS):
                columns.extend(["videos count", "figures count"])
            elif project_info.type in [
                str(sly.ProjectType.POINT_CLOUDS),
                str(sly.ProjectType.POINT_CLOUD_EPISODES),
            ]:
                columns.extend(["pointclouds count", "figures count"])
            elif project_info.type == str(sly.ProjectType.VOLUMES):
                columns.extend(["volumes count", "figures count"])

            class_items = {}
            for item in stats["images"]["objectClasses"]:
                class_items[item["objectClass"]["name"]] = item["total"]

            class_objects = {}
            for item in stats["objects"]["items"]:
                class_objects[item["objectClass"]["name"]] = item["total"]
            for obj_class in data_to_show:
                obj_class["itemsCount"] = class_items[obj_class["title"]]
                obj_class["objectsCount"] = class_objects[obj_class["title"]]

        elif self._project_fs is not None:
            project_stats = self._project_fs.get_classes_stats()

            if type(self._project_fs) == sly.Project:
                columns.extend(["images count", "labels count"])

            elif type(self._project_fs) == sly.VideoProject:
                columns.extend(["videos count", "objects count", "figures count"])

            elif type(self._project_fs) in [sly.PointcloudProject, sly.PointcloudEpisodeProject]:
                columns.extend(["pointclouds count", "objects count", "figures count"])

            elif type(self._project_fs) == sly.VolumeProject:
                columns.extend(["volumes count", "objects count", "figures count"])

            for obj_class in data_to_show:
                obj_class["itemsCount"] = project_stats["items_count"][obj_class["title"]]
                obj_class["objectsCount"] = project_stats["objects_count"][obj_class["title"]]
                if type(self._project_fs) != sly.Project:
                    obj_class["figuresCount"] = project_stats["figures_count"][obj_class["title"]]

        columns = [col.upper() for col in columns]
        if data_to_show:
            table_data = []
            data_to_show = sorted(data_to_show, key=lambda line: line["objectsCount"], reverse=True)
            for line in data_to_show:
                table_line = []
                icon = type_to_zmdi_icon[sly.AnyGeometry]
                for geo_type, icon_text in type_to_zmdi_icon.items():
                    geo_type: Geometry
                    if geo_type.geometry_name() == line["shape"]:
                        icon = icon_text
                        break
                table_line.extend(
                    [
                        {
                            "name": "CLASS",
                            "data": line["title"],
                            "color": line["color"],
                        },
                        {"name": "SHAPE", "data": line["shape"], "icon": icon},
                    ]
                )
                if "itemsCount" in line.keys():
                    table_line.append(
                        {"name": "ITEMS COUNT", "data": line["itemsCount"]}
                    )
                if "objectsCount" in line.keys():
                    table_line.append(
                        {"name": "OBJECTS COUNT", "data": line["objectsCount"]}
                    )
                if "figuresCount" in line.keys():
                    table_line.append(
                        {"name": "FIGURES COUNT", "data": line["figuresCount"]}
                    )
                table_data.append(table_line)
            self._table_data = table_data
            self._columns = columns
            self._checkboxes = [False] * len(table_data)
            self._global_checkbox = False
        else:
            self._table_data = []
            self._columns = []
            self._checkboxes = []
            self._global_checkbox = False

    def read_meta(self, project_meta: sly.ProjectMeta) -> None:
        self.loading = True
        self._project_fs = None
        self._project_id = None
        self.clear_selection()
        self._update_meta(project_meta=project_meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()["checkboxes"] = self._checkboxes
        StateJson()["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def read_project(self, project_fs: sly.Project) -> None:
        self.loading = True
        self._project_fs = project_fs
        self._project_id = None
        self.clear_selection()
        self._update_meta(project_meta=project_fs.meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()["checkboxes"] = self._checkboxes
        StateJson()["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def read_project_from_id(self, project_id: int) -> None:
        self.loading = True
        self._project_fs = None
        self._project_id = project_id
        if self._api is None:
            self._api = sly.Api()
        project_meta = sly.ProjectMeta.from_json(self._api.project.get_meta(project_id))
        self.clear_selection()
        self._update_meta(project_meta=project_meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()["checkboxes"] = self._checkboxes
        StateJson()["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def get_json_data(self) -> Dict[str, Any]:
        return {
            "table_data": self._table_data,
            "columns": self._columns,
            "loading": self._loading,
            "disabled": self._disabled,
            "selectable": self._selectable,
        }

    @property
    def allowed_types(self) -> List[Geometry]:
        return self._allowed_types

    @property
    def project_id(self) -> int:
        return self._project_id

    @property
    def project_fs(self) -> int:
        return self._project_fs

    @property
    def loading(self) -> bool:
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    def get_json_state(self) -> Dict[str, Any]:
        return {
            "global_checkbox": self._global_checkbox,
            "checkboxes": self._checkboxes,
        }

    def get_selected_classes(self) -> List[str]:
        classes = []
        for i, line in enumerate(self._table_data):
            if StateJson()[self.widget_id]["checkboxes"][i]:
                for col in line:
                    if col["name"] == "CLASS":
                        classes.append(col["data"])
        return classes

    def clear_selection(self) -> None:
        self._global_checkbox = False
        self._checkboxes = [False] * len(self._table_data)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()



import os
import shutil
import time

import cv2
import ffmpeg

import supervisely as sly
from tqdm import tqdm


class InferenceVideoInterface:
    def __init__(self, api, start_frame_index, frames_count, frames_direction, video_info, imgs_dir):
        self.api = api

        self.video_info = video_info
        self.images_paths = []

        self.start_frame_index = start_frame_index
        self.frames_count = frames_count
        self.frames_direction = frames_direction

        self._video_fps = round(1 / self.video_info.frames_to_timecodes[1])

        self._geometries = []
        self._frames_indexes = []

        self._add_frames_indexes()

        self._frames_path = os.path.join(imgs_dir, f'video_inference_{video_info.id}_{time.time_ns()}', 'frames')
        self._imgs_dir = imgs_dir

        self._local_video_path = None

        os.makedirs(self._frames_path, exist_ok=True)

    def _add_frames_indexes(self):
        total_frames = self.video_info.frames_count
        cur_index = self.start_frame_index

        if self.frames_direction == 'forward':
            end_point = cur_index + self.frames_count if cur_index + self.frames_count < total_frames else total_frames
            self._frames_indexes = [curr_frame_index for curr_frame_index in range(cur_index, end_point, 1)]
        else:
            end_point = cur_index - self.frames_count if cur_index - self.frames_count > -1 else -1
            self._frames_indexes = [curr_frame_index for curr_frame_index in range(cur_index, end_point, -1)]
            self._frames_indexes = []

    def _download_video_by_frames(self):
        for index, frame_index in tqdm(enumerate(self._frames_indexes), desc='Downloading frames',
                                       total=len(self._frames_indexes)):
            frame_path = os.path.join(f"{self._frames_path}", f"frame{index:06d}.png")
            self.images_paths.append(frame_path)

            if os.path.isfile(frame_path):
                continue

            img_rgb = self.api.video.frame.download_np(self.video_info.id, frame_index)
            cv2.imwrite(os.path.join(f"{self._frames_path}", f"frame{index:06d}.png"),
                        img_rgb)  # save frame as PNG file

    def _download_entire_video(self):
        def videos_to_frames(video_path, frames_range=None):
            def check_rotation(path_video_file):
                # this returns meta-data of the video file in form of a dictionary
                meta_dict = ffmpeg.probe(path_video_file)

                # from the dictionary, meta_dict['streams'][0]['tags']['rotate'] is the key
                # we are looking for
                rotate_code = None
                try:
                    if int(meta_dict['streams'][0]['tags']['rotate']) == 90:
                        rotate_code = cv2.ROTATE_90_CLOCKWISE
                    elif int(meta_dict['streams'][0]['tags']['rotate']) == 180:
                        rotate_code = cv2.ROTATE_180
                    elif int(meta_dict['streams'][0]['tags']['rotate']) == 270:
                        rotate_code = cv2.ROTATE_90_COUNTERCLOCKWISE
                except Exception as ex:
                    pass

                return rotate_code

            def correct_rotation(frame, rotate_code):
                return cv2.rotate(frame, rotate_code)

            vidcap = cv2.VideoCapture(video_path)
            success, image = vidcap.read()
            count = 0
            rotateCode = check_rotation(video_path)

            while success:
                output_image_path = os.path.join(f"{self._frames_path}", f"frame{count:06d}.png")
                if frames_range:
                    if frames_range[0] <= count <= frames_range[1]:
                        if rotateCode is not None:
                            image = correct_rotation(image, rotateCode)
                        cv2.imwrite(output_image_path, image)  # save frame as PNG file
                        self.images_paths.append(output_image_path)
                else:
                    if rotateCode is not None:
                        image = correct_rotation(image, rotateCode)
                    cv2.imwrite(output_image_path, image)  # save frame as PNG file
                    self.images_paths.append(output_image_path)
                success, image = vidcap.read()
                count += 1

            fps = vidcap.get(cv2.CAP_PROP_FPS)

            return {'frames_path': self._frames_path, 'fps': fps, 'video_path': video_path}

        self._local_video_path = os.path.join(self._imgs_dir, f'{time.time_ns()}_{self.video_info.name}')
        self.api.video.download_path(self.video_info.id, self._local_video_path)
        return videos_to_frames(self._local_video_path, [self.start_frame_index, self.start_frame_index + self.frames_count - 1])

    def download_frames(self):
        if self.frames_count > (self.video_info.frames_count * 0.3):
            sly.logger.debug('Downloading entire video')
            self._download_entire_video()
        else:
            sly.logger.debug('Downloading video frame by frame')
            self._download_video_by_frames()

    def __del__(self):
        if os.path.isdir(self._frames_path):
            shutil.rmtree(os.path.dirname(self._frames_path), ignore_errors=True)

            if self._local_video_path is not None and os.path.isfile(self._local_video_path):
                os.remove(self._local_video_path)


# coding: utf-8

from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.api.entity_annotation.figure_api import FigureApi, ApiField


class PointcloudFigureApi(FigureApi):
    def create(self, pointcloud_id, object_id, geometry_json, geometry_type, track_id=None):
        return super().create(pointcloud_id, object_id, {}, geometry_json, geometry_type, track_id)

    def append_bulk(self, pointcloud_id, figures, key_id_map: KeyIdMap):
        keys = []
        figures_json = []
        for figure in figures:
            keys.append(figure.key())
            figures_json.append(figure.to_json(key_id_map))

        self._append_bulk(pointcloud_id, figures_json, keys, key_id_map)

    def append_to_dataset(self, dataset_id, figures, entity_ids, key_id_map: KeyIdMap):
        keys = []
        figures_json = []
        for figure, entity_id in zip(figures, entity_ids):
            keys.append(figure.key())
            figure_json = figure.to_json(key_id_map)
            figure_json[ApiField.ENTITY_ID] = entity_id
            figures_json.append(figure_json)

        return self._append_bulk(dataset_id, figures_json, keys, key_id_map, field_name=ApiField.DATASET_ID)

    def _convert_json_info(self, info: dict, skip_missing=True):
        return super()._convert_json_info(info, skip_missing)




from supervisely.app.widgets import Widget, Select
from typing import Dict


class OneOf(Widget):
    def __init__(
        self,
        conditional_widget: Select,  # or RadioGroup in future
        widget_id: str = None,
    ):
        self._conditional_widget = conditional_widget
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {}

    def get_json_state(self) -> Dict:
        return {}



# coding: utf-8

import os
import urllib
import string
import re
import base64
import hashlib
import json
import numpy as np
import random
from datetime import datetime
import copy
from typing import Optional
from supervisely.io import fs as sly_fs
from supervisely.sly_logger import logger

random.seed(datetime.now())


def rand_str(length):
    chars = string.ascii_letters + string.digits  # [A-z][0-9]
    return "".join((random.choice(chars)) for _ in range(length))


# @TODO: use in API? or remove
def generate_free_name(used_names, possible_name, with_ext=False, extend_used_names=False):
    res_name = possible_name
    new_suffix = 1
    while res_name in set(used_names):
        if with_ext is True:
            res_name = "{}_{:02d}{}".format(
                sly_fs.get_file_name(possible_name),
                new_suffix,
                sly_fs.get_file_ext(possible_name),
            )
        else:
            res_name = "{}_{:02d}".format(possible_name, new_suffix)
        new_suffix += 1
    if extend_used_names:
        used_names.add(res_name)
    return res_name


def generate_names(base_name, count):
    name = sly_fs.get_file_name(base_name)
    ext = sly_fs.get_file_ext(base_name)

    names = [base_name]
    for idx in range(1, count):
        names.append("{}_{:02d}{}".format(name, idx, ext))

    return names


def camel_to_snake(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def take_with_default(v, default):
    return v if v is not None else default


def batched(seq, batch_size=50):
    for i in range(0, len(seq), batch_size):
        yield seq[i : i + batch_size]


def get_bytes_hash(bytes):
    return base64.b64encode(hashlib.sha256(bytes).digest()).decode("utf-8")


def get_string_hash(data):
    return base64.b64encode(hashlib.sha256(str.encode(data)).digest()).decode("utf-8")


def unwrap_if_numpy(x):
    return x.item() if isinstance(x, np.number) else x


def _dprint(json_data):
    print(json.dumps(json_data))


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)


COMMUNITY = "community"
ENTERPRISE = "enterprise"


def validate_percent(value):
    if 0 <= value <= 100:
        pass
    else:
        raise ValueError("Percent has to be in range [0; 100]")


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, "Yi", suffix)


def _remove_sensitive_information(d: dict):
    new_dict = copy.deepcopy(d)
    fields = ["api_token", "API_TOKEN", "AGENT_TOKEN", "apiToken", "spawnApiToken"]
    for field in fields:
        if field in new_dict:
            new_dict[field] = "***"

    for parent_key in ["state", "context"]:
        if parent_key in new_dict and type(new_dict[parent_key]) is dict:
            for field in fields:
                if field in new_dict[parent_key]:
                    new_dict[parent_key][field] = "***"
    return new_dict


def validate_img_size(img_size):
    if not isinstance(img_size, (tuple, list)):
        raise TypeError(
            '{!r} has to be a tuple or a list. Given type "{}".'.format("img_size", type(img_size))
        )
    return tuple(img_size)


def is_development() -> bool:
    mode = os.environ.get("ENV", "development")
    if mode == "production":
        return False
    else:
        return True


def is_debug_with_sly_net() -> bool:
    mode = os.environ.get("DEBUG_WITH_SLY_NET")
    if mode is not None:
        return True
    else:
        return False


def is_production() -> bool:
    return not is_development()


def abs_url(relative_url: str) -> str:
    from supervisely.api.api import SERVER_ADDRESS

    server_address = os.environ.get(SERVER_ADDRESS, "")
    if server_address == "":
        logger.warn("SERVER_ADDRESS env variable is not defined")
    return urllib.parse.urljoin(server_address, relative_url)


def compress_image_url(
    url: str,
    width: Optional[int] = None,
    height: Optional[int] = None,
    quality: Optional[int] = 70,
) -> str:
    if width is None:
        width = ""
    if height is None:
        height = ""
    return url.replace(
        "/image-converter",
        f"/previews/{width}x{height},jpeg,q{quality}/image-converter",
    )


def get_preview_link(title="preview"):
    return (
        f'<a href="javascript:;">{title}<i class="zmdi zmdi-cast" style="margin-left: 5px"></i></a>'
    )


def get_datetime(value: str) -> datetime:
    if value is None:
        return None
    return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")


def get_readable_datetime(value: str) -> str:
    dt = get_datetime(value)
    if dt is None:
        return None
    return dt.strftime("%Y-%m-%d %H:%M:%S")



# coding: utf-8

import os
from collections import defaultdict

from supervisely.io import fs as fs_utils
from supervisely.imaging import image as image_utils
from supervisely.project.project import Project
from supervisely.annotation.annotation import Annotation
from supervisely.geometry.rectangle import Rectangle

OUT_IMG_EXT = '.jpg'
XML_EXT = '.xml'
TXT_EXT = '.txt'


def save_project_as_pascal_voc_segmentation(save_path, project: Project):
    raise NotImplementedError


def save_images_lists(path, tags_to_lists):
    for tag_name, samples_desc_list in tags_to_lists.items():
        with open(os.path.join(path, tag_name + TXT_EXT), 'w') as fout:
            for record in samples_desc_list:
                fout.write('{}  {}\n'.format(record[0], record[1]))  # 0 - sample name, 1 - objects count


def save_project_as_pascal_voc_detection(save_path, project: Project):
    import pascal_voc_writer
    
    # Create root pascal 'datasets' folders
    for dataset in project.datasets:
        pascal_dataset_path = os.path.join(save_path, dataset.name)

        images_dir = os.path.join(pascal_dataset_path, 'JPEGImages')
        anns_dir = os.path.join(pascal_dataset_path, 'Annotations')
        lists_dir = os.path.join(pascal_dataset_path, 'ImageSets/Layout')

        fs_utils.mkdir(pascal_dataset_path)
        for subdir in ['ImageSets',  # Train list, Val list, etc.
                       'ImageSets/Layout',
                       'Annotations',
                       'JPEGImages']:
            fs_utils.mkdir(os.path.join(pascal_dataset_path, subdir))

        samples_by_tags = defaultdict(list)  # TRAIN: [img_1, img2, ..]

        for item_name in dataset:
            img_path, ann_path = dataset.get_item_paths(item_name)
            no_ext_name = fs_utils.get_file_name(item_name)
            pascal_img_path = os.path.join(images_dir, no_ext_name + OUT_IMG_EXT)
            pascal_ann_path = os.path.join(anns_dir, no_ext_name + XML_EXT)


            if item_name.endswith(OUT_IMG_EXT):
                fs_utils.copy_file(img_path, pascal_img_path)
            else:
                img = image_utils.read(img_path)
                image_utils.write(pascal_img_path, img)

            ann = Annotation.load_json_file(ann_path, project_meta=project.meta)

            # Read tags for images lists generation
            for tag in ann.img_tags:
                samples_by_tags[tag.name].append((no_ext_name ,len(ann.labels)))

            writer = pascal_voc_writer.Writer(path=pascal_img_path,
                                              width=ann.img_size[1],
                                              height=ann.img_size[0])

            for label in ann.labels:
                obj_class = label.obj_class
                rect: Rectangle = label.geometry.to_bbox()
                writer.addObject(name=obj_class.name,
                                 xmin = rect.left,
                                 ymin = rect.top,
                                 xmax = rect.right,
                                 ymax = rect.bottom)
            writer.save(pascal_ann_path)

        save_images_lists(lists_dir, samples_by_tags)
